use std::{f32, f64};



// Create the Constants trait so that we can use
// associated constants to define Mathematical constants for
// the f32 and f64 types.
/// Defines Mathematical constants.
pub trait Constants
{
   /// The square root of 2.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::SQRT_2;
   /// let val64: f64 = Constants::SQRT_2;
   /// # assert_eq!(val32, std::f32::consts::SQRT_2);
   /// # assert_eq!(val64, std::f64::consts::SQRT_2);
   ///```
   const SQRT_2: Self;

   /// The square root of 3.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::SQRT_3;
   /// let val64: f64 = Constants::SQRT_3;
   /// # assert_eq!(val32, 1.73205080756887729352f32);
   /// # assert_eq!(val64, 1.73205080756887729352f64);
   ///```
   const SQRT_3: Self;

   /// The inverse of the square root of 2
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::INVERSE_SQRT_2;
   /// let val64: f64 = Constants::INVERSE_SQRT_2;
   /// # assert_eq!(val32, 1.0f32 / std::f32::consts::SQRT_2);
   /// # assert_eq!(val64, 1.0f64 / std::f64::consts::SQRT_2);
   ///```
   const INVERSE_SQRT_2: Self;

   /// The inverse of the square root of 3.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::INVERSE_SQRT_3;
   /// let val64: f64 = Constants::INVERSE_SQRT_3;
   /// # assert_eq!(val32, 1.0f32 / 1.73205080756887729352f32);
   /// # assert_eq!(val64, 1.0f64 / 1.73205080756887729352f64);
   ///```
   const INVERSE_SQRT_3: Self;

   /// The mathematical constant [E][1]. Also known as [Euler's number][1].
   ///
   /// [1]: https://en.wikipedia.org/wiki/E_(mathematical_constant)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::E;
   /// let val64: f64 = Constants::E;
   /// # assert_eq!(val32, std::f32::consts::E);
   /// # assert_eq!(val64, std::f64::consts::E);
   ///```
   const E: Self;

   /// The log base 2 of E.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::LOG2_E;
   /// let val64: f64 = Constants::LOG2_E;
   /// # assert_eq!(val32, std::f32::consts::LOG2_E);
   /// # assert_eq!(val64, std::f64::consts::LOG2_E);
   ///```
   const LOG2_E: Self;

   /// The log base 10 of E.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::LOG10_E;
   /// let val64: f64 = Constants::LOG10_E;
   /// # assert_eq!(val32, std::f32::consts::LOG10_E);
   /// # assert_eq!(val64, std::f64::consts::LOG10_E);
   ///```
   const LOG10_E: Self;

   /// The natural log(ln) of 2.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::LOGE_2;
   /// let val64: f64 = Constants::LOGE_2;
   /// # assert_eq!(val32, 2.0f32.ln());
   /// # assert_eq!(val64, 2.0f64.ln());
   ///```
   const LOGE_2: Self;

   /// The natural log(ln) of 10.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::LOGE_10;
   /// let val64: f64 = Constants::LOGE_10;
   /// # assert_eq!(val32, 10.0f32.ln());
   /// # assert_eq!(val64, 10.0f64.ln());
   ///```
   const LOGE_10: Self;

   /// Two times the value of PI.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::TWO_PI;
   /// let val64: f64 = Constants::TWO_PI;
   /// # assert_eq!(val32, std::f32::consts::PI * 2.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI * 2.0f64);
   ///```
   const TWO_PI: Self;

   /// [PI][1]. The ratio of a circles circumference to its diameter.
   ///
   /// [1]: https://en.wikipedia.org/wiki/Pi
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::PI;
   /// let val64: f64 = Constants::PI;
   /// # assert_eq!(val32, std::f32::consts::PI);
   /// # assert_eq!(val64, std::f64::consts::PI);
   ///```
   const PI: Self;

   /// One half of PI. (PI/2)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::HALF_PI;
   /// let val64: f64 = Constants::HALF_PI;
   /// # assert_eq!(val32, std::f32::consts::PI / 2.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI / 2.0f64);
   ///```
   const HALF_PI: Self;

   /// One third of PI. (PI/3)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::THIRD_PI;
   /// let val64: f64 = Constants::THIRD_PI;
   /// # assert_eq!(val32, std::f32::consts::PI / 3.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI / 3.0f64);
   ///```
   const THIRD_PI: Self;

   /// One forth of PI. (PI/4)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::QUARTER_PI;
   /// let val64: f64 = Constants::QUARTER_PI;
   /// # assert_eq!(val32, std::f32::consts::PI / 4.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI / 4.0f64);
   ///```
   const QUARTER_PI: Self;

   /// One sixth of PI. (PI/6)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::SIXTH_PI;
   /// let val64: f64 = Constants::SIXTH_PI;
   /// # assert_eq!(val32, std::f32::consts::PI / 6.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI / 6.0f64);
   ///```
   const SIXTH_PI: Self;

   /// One eighth of PI. (PI/8)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::EIGHTH_PI;
   /// let val64: f64 = Constants::EIGHTH_PI;
   /// # assert_eq!(val32, std::f32::consts::PI / 8.0f32);
   /// # assert_eq!(val64, std::f64::consts::PI / 8.0f64);
   ///```
   const EIGHTH_PI: Self;

   /// The inverse of PI. (1/PI)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::INVERSE_PI;
   /// let val64: f64 = Constants::INVERSE_PI;
   /// # assert_eq!(val32, 1.0f32 / std::f32::consts::PI);
   /// # assert_eq!(val64, 1.0f64 / std::f64::consts::PI);
   ///```
   const INVERSE_PI: Self;

   /// Two times the inverse of PI. (2/PI)
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::TWO_INVERSE_PI;
   /// let val64: f64 = Constants::TWO_INVERSE_PI;
   /// # assert_eq!(val32, 2.0f32 / std::f32::consts::PI);
   /// # assert_eq!(val64, 2.0f64 / std::f64::consts::PI);
   ///```
   const TWO_INVERSE_PI: Self;

   /// Two times the inverse of the square root of PI. (2/sqrt(PI))
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::TWO_INVERSE_SQRT_PI;
   /// let val64: f64 = Constants::TWO_INVERSE_SQRT_PI;
   /// # assert_eq!(val32, 2.0f32 * (std::f32::consts::PI).sqrt().recip());
   /// # assert_eq!(val64, 2.0f64 * (std::f64::consts::PI).sqrt().recip());
   ///```
   const TWO_INVERSE_SQRT_PI: Self;

   /// PI divided by 180. This is used for radian and degree conversions.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::PI_DIVIDED_BY_180;
   /// let val64: f64 = Constants::PI_DIVIDED_BY_180;
   /// # assert_eq!(val32, (std::f32::consts::PI) / 180.0f32);
   /// # assert_eq!(val64, (std::f64::consts::PI) / 180.0f64);
   ///```
   const PI_DIVIDED_BY_180: Self;

   /// 180 divided by PI. This is used for radian and degree conversions.
   ///
   ///```
   /// use sigils::Constants;
   ///
   /// let val32: f32 = Constants::INVERSE_PI_DIVIDED_BY_180;
   /// let val64: f64 = Constants::INVERSE_PI_DIVIDED_BY_180;
   /// # assert_eq!(val32, 180.0f32 / (std::f32::consts::PI));
   /// # assert_eq!(val64, 180.0f64 / (std::f64::consts::PI));
   ///```
   const INVERSE_PI_DIVIDED_BY_180: Self;
}


// Implement a macro to simplify defining these constants
// for the two different types, f32 and f64.
macro_rules! constants_trait_impl
{
   ($T: ty, $sqrt2: expr, $sqrt3: expr, $e: expr,
    $log2e: expr, $log10e: expr, $loge2: expr, $loge10: expr,
    $pi: expr, $sqrtPI: expr) =>
   {
      impl Constants for $T
      {
         const SQRT_2: $T = $sqrt2;
         const SQRT_3: $T = $sqrt3;
         const INVERSE_SQRT_2: $T = 1.0 / $sqrt2;
         const INVERSE_SQRT_3: $T = 1.0 / $sqrt3;

         const E: $T = $e;

         const LOG2_E: $T = $log2e;
         const LOG10_E: $T = $log10e;
         const LOGE_2: $T = $loge2;
         const LOGE_10: $T = $loge10;

         const TWO_PI: $T = 2.0 * $pi;
         const PI: $T = $pi;
         const HALF_PI: $T = $pi / 2.0;
         const THIRD_PI: $T = $pi / 3.0;
         const QUARTER_PI: $T = $pi / 4.0;
         const SIXTH_PI: $T = $pi / 6.0;
         const EIGHTH_PI: $T = $pi / 8.0;
         const INVERSE_PI: $T = 1.0 / $pi;
         const TWO_INVERSE_PI: $T = 2.0 / $pi;
         const TWO_INVERSE_SQRT_PI: $T = 2.0 / $sqrtPI;
         const PI_DIVIDED_BY_180: $T = $pi / 180.0;
         const INVERSE_PI_DIVIDED_BY_180: $T = 180.0 / $pi;
      }
   }
}


// Implement the math Constants.
constants_trait_impl!(f32, f32::consts::SQRT_2, 1.73205080756887729352f32,
                      f32::consts::E, f32::consts::LOG2_E,
                      f32::consts::LOG10_E, f32::consts::LN_2,
                      f32::consts::LN_10, f32::consts::PI,
                      1.7724538509055159f32);
constants_trait_impl!(f64, f64::consts::SQRT_2, 1.73205080756887729352f64,
                      f64::consts::E, f64::consts::LOG2_E,
                      f64::consts::LOG10_E, f64::consts::LN_2,
                      f64::consts::LN_10, f64::consts::PI,
                      1.7724538509055159f64);
