use std::ops::{Add, Sub, Mul, Div, Rem, Neg};
use std::ops::{AddAssign, SubAssign, MulAssign, DivAssign, RemAssign};

use crate::constants::Constants;
use crate::real::Real;
use crate::trig::degree::Degree;



/// A unit of angle, equal to an angle at the center of
/// a circle whose arc is equal in length to the radius.
#[derive(Clone, Copy, Eq, Ord, PartialEq, PartialOrd, Hash)]
pub struct Radian<T>
{
   value: T
}



impl<T> Radian<T> where T: Real
{
   /// Create a new Radian structure with the given value.
   pub fn new(val: T) -> Self
   {
      Radian
      {
         value: val
      }
   }
}

impl<T> ::std::ops::Deref for Radian<T> where T: Real 
{
   type Target = T;

   fn deref(&self) -> &T
   {
      &self.value
   }
}

impl<T> ::std::ops::DerefMut for Radian<T> where T: Real 
{
   fn deref_mut<'a>(&'a mut self) -> &'a mut T
   {
      &mut self.value
   }
}

binary_operator_impl!(Add::add, Real, Radian {value});
binary_operator_impl!(Sub::sub, Real, Radian {value});
binary_operator_impl!(Mul::mul, Real, Radian {value});
binary_operator_impl!(Div::div, Real, Radian {value});
binary_operator_impl!(Rem::rem, Real, Radian {value});

impl<T> ::std::ops::Neg for Radian<T>
   where T: Neg<Output = T>
{
   type Output = Radian<T>;

   fn neg(self) -> Radian<T>
   {
      let mut radian: Radian<T>;

      radian = self;
      radian.value = radian.value.neg();
      radian
   }
}

binary_operator_assign_impl!(AddAssign::add_assign,
                             Real, Radian {value});
binary_operator_assign_impl!(SubAssign::sub_assign,
                             Real, Radian {value});
binary_operator_assign_impl!(MulAssign::mul_assign,
                             Real, Radian {value});
binary_operator_assign_impl!(DivAssign::div_assign,
                             Real, Radian {value});
binary_operator_assign_impl!(RemAssign::rem_assign,
                             Real, Radian {value});

impl ::std::convert::From<Radian<f32>> for f32
{
   fn from(radian: Radian<f32>) -> f32
   {
      radian.value
   }
}

impl ::std::convert::From<Radian<f64>> for f64
{
   fn from(radian: Radian<f64>) -> f64
   {
      radian.value
   }
}

impl<T> ::std::convert::From<T> for Radian<T> where T: Real 
{
   fn from(val: T) -> Radian<T>
   {
      Radian::new(val)
   }
}

impl<T> ::std::convert::From<Degree<T>> for Radian<T>
   where T: Real 
{
   fn from(degrees: Degree<T>) -> Radian<T>
   {
      let rads: T;

      rads = *degrees * Constants::PI_DIVIDED_BY_180;
      Radian::new(rads)
   }
}

impl<T> ::std::fmt::Debug for Radian<T>
   where T: Real 
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      ::std::fmt::Display::fmt(self, f)
   }
}

impl<T> ::std::fmt::Display for Radian<T>
   where T: Real 
{
   fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result
   {
      write!(f, "Radian({})", self.value)
   }
}
