use std::ops::Add;

/// Defines an additive identity element for `Self`.
pub trait Zero: Sized + Add<Self, Output=Self>
{
   /// Returns the additive identity element of `Self`, `0`.
   ///
   /// # Laws
   ///
   /// ```text
   /// a + 0 = a       ∀ a ∈ Self
   /// 0 + a = a       ∀ a ∈ Self
   /// ```
   ///
   /// # Purity
   ///
   /// This function should return the same result at all times regardless of
   /// external mutable state, for example values stored in TLS or in
   /// `static mut`s.
   fn zero() -> Self;
}



macro_rules! zero_impl
{
   ($varType: ty, $val: expr) =>
   {
      impl Zero for $varType
      {
         fn zero() -> $varType
         {
            $val
         }
      }
   }
}



zero_impl!(u8, 0u8);
zero_impl!(u16, 0u16);
zero_impl!(u32, 0u32);
zero_impl!(u64, 0u64);
zero_impl!(usize, 0usize);

zero_impl!(i8, 0i8);
zero_impl!(i16, 0i16);
zero_impl!(i32, 0i32);
zero_impl!(i64, 0i64);
zero_impl!(isize, 0isize);

zero_impl!(f32, 0.0f32);
zero_impl!(f64, 0.0f64);
